/*
 * Read a file in RLE life format (run-length encoded).
 *
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "hashlife.h"

#define	isBlank(ch)	(((ch) == ' ') || ((ch) == '\t'))
#define	isDigit(ch)	(((ch) >= '0') && ((ch) <= '9'))


/*
 * Read the specified file name which contains a Life object in run-length-
 * encoded format, and add it to the current view of the specified object.
 * Returns FALSE with a message printed if the file cannot be read or is
 * in the wrong format.
 */
BOOL
readRle(const char * name, OBJ * obj)
{
	FILE *	fp;
	long	line = 0;
	const char * cp;
	NUMBER	num;
	NUMBER	minX;
	NUMBER	curX;
	NUMBER	curY;
	int	ch;
	BOOL	sawNum;
	char	buf[256];

	fp = fopen(name, "r");

	if (fp == 0)
	{
		fprintf(stderr, "Cannot open \"%s\" for reading: %s\n",
			name, strerror(errno));

		return FALSE;
	}

	/*
	 * Look for the first line in the following format:
	 *	x = <number>, y = <number>, rule=<ruleString>
	 * which give the sizes for the x and y dimensions.
	 * The rule string is optional.  Ignore blank lines and
	 * lines beginning with hash marks.
	 */
	do
	{
		int len;

		if (fgets(buf, sizeof(buf), fp) == 0)
		{
			fprintf(stderr, "Missing RLE definition line\n");

			fclose(fp);

			return FALSE;
		}

		line++;

		/*
		 * Remove any trailing newline and blanks.
		 */
		len = strlen(buf);

		if ((len > 0) && (buf[len - 1] == '\n'))
			len--;

		while ((len > 0) && isBlank(buf[len - 1]))
			len--;

		buf[len] = '\0';

		/*
		 * Skip blanks and check the next character.
		 * If the line is blank or a comment then go on.
		 */
		cp = buf;

		while (isBlank(*cp))
			cp++;
	}
	while ((*cp == '\0') || (*cp == '#'));

	/*
	 * Now this should be the RLE definition line.
	 * Start with 'x = <num>'.
	 */
	if ((*cp != 'x') && (*cp != 'X'))
		goto badFormat;

	cp++;

	while (isBlank(*cp))
		cp++;

	if (*cp++ != '=')
		goto badFormat;

	while (isBlank(*cp))
		cp++;

	if (!isDigit(*cp))
		goto badFormat;

	num = numZero;

	while (isDigit(*cp))
	{
		num = numMulLong(num, 10);
		num = numAddLong(num, *cp++ - '0');
	}

	curX = numHalf(num);
	numNegAddr(&curX);
	minX = curX;

	/*
	 * Now check for 'y = <num>' following a comma.
	 */
	while (isBlank(*cp))
		cp++;

	if (*cp++ != ',')
		goto badFormat;

	while (isBlank(*cp))
		cp++;

	if ((*cp != 'y') && (*cp != 'Y'))
		goto badFormat;

	cp++;

	while (isBlank(*cp))
		cp++;

	if (*cp++ != '=')
		goto badFormat;

	while (isBlank(*cp))
		cp++;

	if (!isDigit(*cp))
		goto badFormat;

	num = numZero;

	while (isDigit(*cp))
	{
		num = numMulLong(num, 10);
		num = numAddLong(num, *cp++ - '0');
	}

	curY = numHalf(num);
	numNegAddr(&curY);

	/*
	 * Then look for the optional rule string following a comma.
	 */
	while (isBlank(*cp))
		cp++;

	if (*cp == ',')
	{
		cp++;

		while (isBlank(*cp))
			cp++;

		if (memcmp(cp, "rule", 4) != 0)
			goto badFormat;

		cp += 4;

		if (!isBlank(*cp))
			goto badFormat;

		while (isBlank(*cp))
			cp++;

		if (*cp++ != '=')
			goto badFormat;

		while (isBlank(*cp))
			cp++;

		if (*cp == '\0')
			goto badFormat;

		if (!mcSetRule(cp))
		{
			fprintf(stderr, "Bad rule in RLE file\n");

			fclose(fp);

			return FALSE;
		}

		cp = "";
	}

	/*
	 * There must be nothing left on the line.
	 */
	if (*cp)
		goto badFormat;

	/*
	 * Ok, the definition line is parsed.
	 * Now look for sequences which are the following:
	 *	<number>b
	 *	<number>o
	 *	<number>$
	 *	!
	 * which are blanks, on cells, newlines, and end of object.
	 * Ignore blanks and end of line characters.
	 */
	sawNum = FALSE;

	while (TRUE)
	{
		ch = fgetc(fp);

		switch (ch)
		{
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				if (!sawNum)
					num = numZero;

				sawNum = TRUE;
				num = numMulLong(num, 10);
				num = numAddLong(num, ch - '0');

				break;

			case 'O':
			case 'o':
				if (!sawNum)
					num = numOne;

				sawNum = FALSE;
				
				while (isPos(num))
				{
					setCell(obj, curX, curY);

					numDecAddr(&num);
					numIncAddr(&curX);
				}

				break;

			case 'B':
			case 'b':
				if (!sawNum)
					num = numOne;

				sawNum = FALSE;

				numAddAddr(&curX, &num);

				break;

			case '$':
				if (!sawNum)
					num = numOne;

				sawNum = FALSE;

				numAddAddr(&curY, &num);
				curX = minX;

				break;

			case ' ':
			case '\t':
				if (sawNum)
					goto badFormat;

				break;

			case '\n':
				if (sawNum)
					goto badFormat;

				line++;

				break;

			case '!':
				if (sawNum)
					goto badFormat;

				fclose(fp);

				return TRUE;

			case EOF:
				fprintf(stderr, "Missing exclamation mark\n");

				fclose(fp);

				return FALSE;

			default:
				goto badFormat;
		}
	}


	/*
	 * Error handling for illegal formats.
	 */
badFormat:
	fprintf(stderr, "Bad RLE format in \"%s\" at line %ld\n", name, line);

	fclose(fp);

	return FALSE;
}
